import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;
import java.awt.image.BufferedImage;
import java.io.File;
import java.io.IOException;

import javax.imageio.ImageIO;
import javax.swing.BoxLayout;
import javax.swing.JFrame;
import javax.swing.JLabel;
import javax.swing.JMenu;
import javax.swing.JMenuBar;
import javax.swing.JMenuItem;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.UIManager;
import javax.swing.UnsupportedLookAndFeelException;
import javax.swing.border.EtchedBorder;
import javax.swing.plaf.metal.MetalLookAndFeel;


public class Rahmenwerk {
	
	public static interface Algorithm {
		public void start(int threshold);
	}
	
	private static String IMAGEPATH = "." + File.separator + "Hand005.jpg";

	private Algorithm algo;
	
	private boolean ready;
	
	private int threshold;
	
	private BufferedImage originalImage;
	private BufferedImage testedImage;
	private BufferedImage markedImage;
	private int mark;
	private int width;
	private int height;	
	
	private JFrame mainFrame;		
	private JLabel statusLine;	
	private JMenuBar menuBar;
	private ImageCanvas originalImageCanvas;
	private ImageCanvas otherImageCanvas;
	
	

	/** 
	 * Creates a Framework + Gui.
	 */
	public Rahmenwerk() {
		this(null);
	}
	
	/** 
	 * Creates a Framework + Gui for the given algorithm.
	 */
	public Rahmenwerk(Algorithm algo) {
		
		this.algo = algo;
		
		this.ready = false;
		
		this.threshold = 150;
				
		try {
			UIManager.setLookAndFeel(new MetalLookAndFeel());
		} catch (UnsupportedLookAndFeelException e) {
			System.err.println("error while setting look-and-feel");
		}		
		this.mainFrame = new JFrame();
		
        javax.swing.SwingUtilities.invokeLater(new Runnable() {
            public void run() {
        		init();
            }
        });
    	mainFrame.setSize(800, 600);
    	mainFrame.setVisible(true);
	}

	/** 
	 * Reloades the image.
	 */
	public synchronized void reset() {
		try {if (!this.ready) wait();} catch (Exception ex) {}
		this.markedImage = loadImage();
		resetImages();
	}

	/** 
	 * The image's width.
	 */
	public synchronized int getWidth() {
		try {if (!this.ready) wait();} catch (Exception ex) {}
		return this.width;
	}

	/** 
	 * The image's height.
	 */
	public synchronized int getHeight() {
		try {if (!this.ready) wait();} catch (Exception ex) {}
		return this.height;
	}

	/** 
	 * Lightness of pixel x, y.
	 */
	public synchronized int getPixel(int x, int y) {
		try {if (!this.ready) wait();} catch (Exception ex) {}
		if (x < 0 || x >= this.width || y < 0 || y >= this.height) return 0;
		int pixel = this.testedImage.getRGB(x, y);
		int h1 = pixel & 0xFF;
		int h2 = (pixel & 0xFF00) >> 8 ;
		int h3 = (pixel & 0xFF0000) >> 16;
		
		//return pixel == -16777216 ? 0 : 200;
		return (h1 + h2 + h3) / 3; 
	}

	/** 
	 * Marks pixel x, y with color.
	 */
	public synchronized void markPixel(int x, int y) {
		try {if (!this.ready) wait();} catch (Exception ex) {}
		if (x < 0 || x >= this.width || y < 0 || y >= this.height) return;
		this.markedImage.setRGB(x, y, mark);
	}

	/** 
	 * Repaints the images.
	 */
	public synchronized void repaint() {
		try {if (!this.ready) wait();} catch (Exception ex) {}
		otherImageCanvas.paintImmediately(0, 0, otherImageCanvas.getWidth(), otherImageCanvas.getHeight());
	}
	
	/** initializes this object */
	private void init() {
		createWindowListener();
		createLayout();
		createMenu();
		createStatusBar();
		
		String title = "Segmentation" + ((algo == null) ? "" : (" with '" + algo.getClass().getName() + "'"));
		setTitle(title);
		
		this.originalImage = loadImage();
		if (this.originalImage != null) {
			this.width = this.originalImage.getWidth();
			this.height= this.originalImage.getHeight();
			this.testedImage = copyImage(this.originalImage);
			this.markedImage = copyImage(this.originalImage);
		}
		this.mark = Color.red.getRGB();

		resetImages();
		
		synchronized(this) {
			this.ready = true;
			this.notifyAll();
		}

	}
	
	private void setTitle(String title) {
		mainFrame.setTitle(title);
	}
	
	private void stop() {
		mainFrame.setVisible(false);		
	}

	/**
	 * creates and attaches the statusbar
	 */
	private void createStatusBar() {
		JPanel statusPanel = new JPanel();
		statusPanel.setBorder(new EtchedBorder());
		statusPanel.setLayout(new BorderLayout());
		this.statusLine = new JLabel("   ");
		statusPanel.add(this.statusLine, BorderLayout.CENTER);
		mainFrame.getContentPane().add(statusPanel, BorderLayout.SOUTH);
	}


	/**
	 * creates and activates the window listener for this MainFrame
	 */
	private void createWindowListener() {
		mainFrame.addWindowListener(new WindowAdapter() {
			public void windowClosing(WindowEvent e) {
				com_quit();
			}
		});
	}

	/**
	 * creates the menu
	 */
	private void createMenu() {
		this.menuBar = new JMenuBar();
		{
			JMenu menuFile = new JMenu("File");
						
			// reset
			JMenuItem itemReset = new JMenuItem("Reset");
			itemReset.addActionListener(new ActionListener() {
				public void actionPerformed(ActionEvent e) {
					Rahmenwerk.this.com_reset();
				}
			});
			if (algo == null) {
				itemReset.setEnabled(false);
			}
			menuFile.add(itemReset);
			
			// go
			JMenuItem itemGo = new JMenuItem("Go");
			itemGo.addActionListener(new ActionListener() {
				public void actionPerformed(ActionEvent e) {
					Rahmenwerk.this.com_go();
				}
			});
			if (algo == null) {
				itemGo.setEnabled(false);
			}
			menuFile.add(itemGo);
			
			// separator
			menuFile.addSeparator();
			
			// quit
			JMenuItem itemQuit = new JMenuItem("Exit");
			itemQuit.addActionListener(new ActionListener() {
				public void actionPerformed(ActionEvent e) {
					Rahmenwerk.this.com_quit();
				}
			});
			menuFile.add(itemQuit);
			
			this.menuBar.add(menuFile);
		}		
		mainFrame.setJMenuBar(this.menuBar);
	}
	
	
	private void resetImages() {				
		originalImageCanvas.setImage(this.testedImage);
		otherImageCanvas.setImage(this.markedImage);		
	}

	private BufferedImage loadImage() {
		BufferedImage image = null;
		try {
			image = ImageIO.read(new File(IMAGEPATH));
		} catch (IOException ex) {
			showStatus("image not available");
		}
		return image;
	}

	private BufferedImage copyImage(BufferedImage oldImage) {
		int width = oldImage.getWidth();
		int height = oldImage.getHeight();
		BufferedImage newImage = new BufferedImage(width, height, BufferedImage.TYPE_INT_ARGB);
		for (int r = 0; r < height; r++) {
			for (int c = 0; c < width; c++) {
				newImage.setRGB(c, r, oldImage.getRGB(c, r));
			}
		}
		return newImage;
	}
		
	/**
	 * creates and activates the layout
	 */
	private void createLayout() {
		mainFrame.getContentPane().setLayout(new BorderLayout());
		mainFrame.getContentPane().add(new JLabel(" "), BorderLayout.NORTH);

		JPanel mainPanel = new JPanel();
		mainPanel.setLayout(new BoxLayout(mainPanel, BoxLayout.X_AXIS));

		originalImageCanvas = new ImageCanvas(this.testedImage, false);
		mainPanel.add(originalImageCanvas);

		otherImageCanvas = new ImageCanvas(this.markedImage, false);
		mainPanel.add(otherImageCanvas);	

		mainFrame.getContentPane().add(mainPanel, BorderLayout.CENTER);
		
	}

	/**
	 * updates the status bar
	 */
	private void showStatus(String message) {
		statusLine.setText(message);
	}
	
	private void com_quit() {
		this.stop();
		System.out.println("...terminated");
		System.exit(0);
	}

	private void com_reset() {
		showStatus("reset");
		reset();
	}

	private void com_go() {
		showStatus("go");
		reset();
		
		String str = JOptionPane.showInputDialog(mainFrame, "Threshold:", new Integer(threshold));
		if (str == null) {
			showStatus("input aborted");
			return;
		}
		try {
			threshold = Integer.parseInt(str);
		} catch (Exception ex) {
			showStatus("input not a legal threshold");
			return;
		}
		if (algo != null) {
			algo.start(threshold);
		}
		
	}	
}
